@@ -19,16 +19,16 @@ typedef struct _GtkImageToolPainterClass GtkImageToolPainterClass;
 
 struct _GtkImageToolPainter
 {
-    GObject          parent;
-    GtkImageView    *view;
+    GObject             parent;
+    GtkImageView       *view;
 
     /* Cursor to use */
-    GdkCursor       *crosshair;
+    GdkCursor          *crosshair;
 
-    /* Drawer for the image */
-    ImageViewDrawer *drawer;
+    /* Cache for the image */
+    GdkPixbufDrawCache *cache;
 
-    MouseHandler    *mouse_handler;
+    MouseHandler       *mouse_handler;
 };
 
 struct _GtkImageToolPainterClass
