@@ -211,42 +211,6 @@ hotspot_list_find_hotspot (Hotspot hotspots[HOTSPOT_LAST],
 /*************************************************************/
 /***** Static stuff ******************************************/
 /*************************************************************/
-/**
- * gtk_image_tool_selector_image_to_widget_rect:
- * @rect_out: a #GdkRectangle to fill in with the widget space
- *   coordinates.
- * @returns: %TRUE if the conversion was successful, %FALSE otherwise.
- *
- * Convert a rectangle in image space coordinates to widget space
- * coordinates. If the view is not realized, or if it contains no
- * pixbuf, then the conversion was unsuccessful, %FALSE is returned
- * and @rect_out is left unmodified.
- **/
-static gboolean
-gtk_image_view_image_to_widget_rect (GtkImageView *view,
-                                     GdkRectangle *rect_in,
-                                     GdkRectangle *rect_out)
-{
-    gdouble zoom = gtk_image_view_get_zoom (view);
-    GdkRectangle zoom_rect = {
-        rect_in->x * zoom,
-        rect_in->y * zoom,
-        rect_in->width * zoom,
-        rect_in->height * zoom
-    };
-
-    GdkRectangle image_rect, viewport;
-    if (!gtk_image_view_get_draw_rect (view, &image_rect))
-        return FALSE;
-    
-    gtk_image_view_get_viewport (view, &viewport);
-    rect_out->x = image_rect.x + zoom_rect.x - viewport.x;
-    rect_out->y = image_rect.y + zoom_rect.y - viewport.y;
-    rect_out->width = zoom_rect.width;
-    rect_out->height = zoom_rect.height;
-    return TRUE;
-}
-
 /**
  * gtk_image_tool_selector_update_selection:
  *
@@ -422,6 +386,8 @@ cursor_at_point (GtkIImageTool *tool,
     GtkImageToolSelector *selector = GTK_IMAGE_TOOL_SELECTOR (tool);
     HotspotType type =
         gtk_image_tool_selector_get_hotspot_at_point (selector, x, y);
+    if (type == HOTSPOT_LAST)
+        return NULL;
     return selector->hotspots[type].cursor;
 }
 
@@ -534,13 +500,10 @@ motion_notify (GtkIImageTool  *tool,
 
 static void
 pixbuf_changed (GtkIImageTool *tool,
-                gboolean       reset_fit)
+                gboolean       reset_fit,
+                GdkRectangle  *rect)
 {
     GtkImageToolSelector *selector = GTK_IMAGE_TOOL_SELECTOR (tool);
-    if (selector->background)
-        g_object_unref (selector->background);
-    selector->background = NULL;
-
     if (reset_fit)
         selector->sel_rect = (GdkRectangle){0, 0, 0, 0};
 
@@ -548,48 +511,65 @@ pixbuf_changed (GtkIImageTool *tool,
     if (!pixbuf)
         return;
 
-    // Update the background.
-    selector->background = gdk_pixbuf_copy (pixbuf);
-    gdk_pixbuf_shade (selector->background);
+    if (rect)
+    {
+        // Copy the damaged area from the foreground to the
+        // background.
+        gdk_pixbuf_copy_area (pixbuf,
+                              rect->x, rect->y,
+                              rect->width, rect->height,
+                              selector->background,
+                              rect->x, rect->y);
+    }
+    else
+    {
+        if (selector->background)
+            g_object_unref (selector->background);
+        selector->background = gdk_pixbuf_copy (pixbuf);
+    }
+    // Update the relevant area of the background.
+    gdk_pixbuf_shade (selector->background, rect);
 
     // Clear caches
-    image_view_drawer_force_scale (selector->bg_drawer);
-    image_view_drawer_force_scale (selector->fg_drawer);
+    gdk_pixbuf_draw_cache_invalidate (selector->bg_cache);
+    gdk_pixbuf_draw_cache_invalidate (selector->fg_cache);
 }
 
 static void
-paint_image (GtkIImageTool *tool,
-             DrawSettings  *ds,
-             GdkDrawable   *drawable)
+paint_image (GtkIImageTool     *tool,
+             GdkPixbufDrawOpts *opts,
+             GdkDrawable       *drawable)
 {
     GtkImageToolSelector *selector = GTK_IMAGE_TOOL_SELECTOR (tool);
 
     // Draw the shaded background.
-    DrawSettings bg_ds = *ds;
-    bg_ds.pixbuf = selector->background;
-    image_view_drawer_draw (selector->bg_drawer, &bg_ds, drawable);
+    GdkPixbufDrawOpts bg_opts = *opts;
+    bg_opts.pixbuf = selector->background;
+    gdk_pixbuf_draw_cache_draw (selector->bg_cache, &bg_opts, drawable);
 
     // Draw the selected area.
     // Intersect the selection rectangle with the exposed rectangle.
     GdkRectangle zoom_sel_rect = {
-        selector->sel_rect.x * ds->zoom,
-        selector->sel_rect.y * ds->zoom,
-        selector->sel_rect.width * ds->zoom,
-        selector->sel_rect.height * ds->zoom
+        selector->sel_rect.x * opts->zoom,
+        selector->sel_rect.y * opts->zoom,
+        selector->sel_rect.width * opts->zoom,
+        selector->sel_rect.height * opts->zoom
     };
     GdkRectangle zoom_paint_rect;
     if (!gdk_rectangle_intersect (&zoom_sel_rect,
-                                  &ds->zoom_rect,
+                                  &opts->zoom_rect,
                                   &zoom_paint_rect))
         return;
 
-    DrawSettings fg_ds = *ds;
+    GdkPixbufDrawOpts fg_opts = *opts;
 
-    fg_ds.zoom_rect = zoom_paint_rect;
-    fg_ds.widget_x = ds->widget_x + zoom_paint_rect.x - ds->zoom_rect.x;
-    fg_ds.widget_y = ds->widget_y + zoom_paint_rect.y - ds->zoom_rect.y;
+    fg_opts.zoom_rect = zoom_paint_rect;
+    fg_opts.widget_x =
+        opts->widget_x + zoom_paint_rect.x - opts->zoom_rect.x;
+    fg_opts.widget_y =
+        opts->widget_y + zoom_paint_rect.y - opts->zoom_rect.y;
     
-    image_view_drawer_draw (selector->fg_drawer, &fg_ds, drawable);
+    gdk_pixbuf_draw_cache_draw (selector->fg_cache, &fg_opts, drawable);
 
     // Draw the selection rectangle.
     GdkGC *rect_gc = gdk_gc_new (drawable);
@@ -637,8 +617,8 @@ gtk_image_tool_selector_finalize (GObject *object)
     GtkImageToolSelector *selector = GTK_IMAGE_TOOL_SELECTOR (object);
     if (selector->background)
         g_object_unref (selector->background);
-    image_view_drawer_free (selector->bg_drawer);
-    image_view_drawer_free (selector->fg_drawer);
+    gdk_pixbuf_draw_cache_free (selector->bg_cache);
+    gdk_pixbuf_draw_cache_free (selector->fg_cache);
 
     gdk_cursor_unref (selector->drag_cursor);
     hotspot_list_free (selector->hotspots);
@@ -737,8 +717,8 @@ gtk_image_tool_selector_init (GtkImageToolSelector *tool)
     tool->background = NULL;
     tool->view = NULL;
     tool->sel_rect = (GdkRectangle){0, 0, 0, 0};
-    tool->bg_drawer = image_view_drawer_new ();
-    tool->fg_drawer = image_view_drawer_new ();
+    tool->bg_cache = gdk_pixbuf_draw_cache_new ();
+    tool->fg_cache = gdk_pixbuf_draw_cache_new ();
 
     tool->drag_cursor = cursor_get (CURSOR_HAND_CLOSED);
     tool->mouse_handler = mouse_handler_new (tool->drag_cursor);
@@ -797,8 +777,8 @@ gtk_image_tool_selector_get_selection (GtkImageToolSelector *selector,
  * @rect: Selection rectangle in image space coordinates.
  *
  * Sets the selection rectangle for the tool. Setting this attribute
- * will cause the widget to immidiately repaint itself if its view is
- * realized. 
+ * will cause the widget to immediately repaint itself if its view is
+ * realized.
  *
  * This method does nothing under the following circumstances:
  *
