@@ -188,17 +188,17 @@
  *     behaviour is most often beneficial and makes the widget very
  *     fast. For example, try opening a very large image (4000x2000
  *     pixels or so) in GtkImageView. The widget will spend some time
- *     bilinearly downsampling the image at the start. Then try
- *     minimizing and unminimizing the window. The image will reappear
- *     immidately because the view has cached it.
+ *     bilinearly scaling the image at the start. Then try minimizing
+ *     and unminimizing the window. The image will reappear immedately
+ *     because the view has cached it.
  *   </para>
  *   <para>
  *     However, this feature means that a client application must
- *     signal to the view when it changes the pixels on the pixbuf
- *     that the view shows. The right way to do that is to use the
- *     gtk_image_view_set_pixbuf() function with the old pixbuf as the
- *     argument. In particular, code that merely tries to update the
- *     view by requesting that it should be redrawn will not work.
+ *     signal to the view when it changes the pixels on the pixbuf the
+ *     view shows. The right way to do that is to use the
+ *     gtk_image_view_damage_pixels() function. Code that merely tries
+ *     to update the view by requesting that it should be redrawn will
+ *     not work.
  *   </para>
  *   <informalexample>
  *     <programlisting>
@@ -206,11 +206,6 @@
  *       gtk_widget_queue_draw_area (10, 10, 50, 50) // Incorrect!
  *     </programlisting>
  *   </informalexample>
- *   <para>
- *     In future versions of GtkImageView a method will be added that
- *     marks an area of the pixbuf as dirty and forces the view to
- *     flush its cache.
- *   </para>
  * </refsect2>
  * <refsect1>
  *   <title>Example</title>
@@ -367,9 +362,10 @@ gtk_image_view_clamp_offset (GtkImageView *view, gdouble *x, gdouble *y)
  * use that cursor.
  *
  * Consider for example if the pointer is over a hotspot in the image.
- * If a key is pressed, then the hotspot is moved, but the cursor is
- * not so no motion event is fired.  Therefore, this method must be
- * called BOTH when the mouse is moved and when the offset is changed.
+ * If an arrow key is pressed, then the hotspot is moved, but the
+ * cursor is not so no motion event is fired. Therefore, this method
+ * must be called BOTH when the mouse is moved and when the offset is
+ * changed.
  **/
 static void
 gtk_image_view_update_cursor (GtkImageView *view)
@@ -521,12 +517,19 @@ gtk_image_view_repaint_area (GtkImageView *view,
     // Do not draw zero size rectangles.
     if (!paint_rect->width || !paint_rect->height)
         return FALSE;
-    view->is_rendering = TRUE;
 
+    // This is a workaround for #80925. If the image area gets to
+    // small, then gdk-pixbuf's scale matrix becomes ridiculously
+    // large which causes noticable and wasteful cpu spikes.
+    Size alloc = gtk_image_view_get_allocated_size (view);
+    if (alloc.width <= 1 || alloc.height <= 1)
+        return FALSE;
+    
+    view->is_rendering = TRUE;
+    
     // Image area is the area on the widget occupied by the pixbuf. 
     GdkRectangle image_area;
     gtk_image_view_get_draw_rect (view, &image_area);
-    Size alloc = gtk_image_view_get_allocated_size (view);
     if (image_area.x > 0 ||
         image_area.y > 0 ||
         image_area.width < alloc.width ||
@@ -556,16 +559,17 @@ gtk_image_view_repaint_area (GtkImageView *view,
         int src_x = view->offset_x + paint_area.x - image_area.x;
         int src_y = view->offset_y + paint_area.y - image_area.y;
 
-        DrawSettings ds = {
+        GdkPixbufDrawOpts opts = {
             view->zoom,
-            (GdkRectangle){src_x, src_y, paint_area.width, paint_area.height},
+            (GdkRectangle){src_x, src_y,
+                           paint_area.width, paint_area.height},
             paint_area.x, paint_area.y,
             interp,
             view->pixbuf,
             view->check_color1,
             view->check_color2
         };
-        gtk_iimage_tool_paint_image (view->tool, &ds, widget->window);
+        gtk_iimage_tool_paint_image (view->tool, &opts, widget->window);
     }
 
     view->is_rendering = FALSE;
@@ -672,7 +676,7 @@ gtk_image_view_fast_scroll (GtkImageView *view,
  * @set_adjustments: whether to update the adjustments. Because this
  *   function is called from the adjustments callbacks, it needs to be
  *   %FALSE to prevent infinite recursion.
- * @invalidate: whether to invalidate the view or redraw immidately,
+ * @invalidate: whether to invalidate the view or redraw immedately,
  *  see gtk_image_view_set_offset()
  *
  * Set the offset of where in the image the #GtkImageView should begin
@@ -1030,7 +1034,7 @@ gtk_image_view_init_signals (GtkImageViewClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__DOUBLE,
                       G_TYPE_NONE,
-                      1, GTK_TYPE_DOUBLE);
+                      1, G_TYPE_DOUBLE);
     gtk_image_view_signals[ZOOM_IN] =
         g_signal_new ("zoom_in",
                       G_TYPE_FROM_CLASS (klass),
@@ -1074,7 +1078,7 @@ gtk_image_view_init_signals (GtkImageViewClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__BOOLEAN,
                       G_TYPE_NONE,
-                      1, GTK_TYPE_INT);
+                      1, G_TYPE_INT);
     /**
      * GtkImageView::scroll:
      * @view: The #GtkImageView that received the signal.
@@ -1158,9 +1162,10 @@ gtk_image_view_init_signals (GtkImageViewClass *klass)
      * @view: The view that emitted the signal.
      *
      * The ::pixbuf-changed signal is emitted when the pixbuf the
-     * image view shows is changed. Listening to this signal is useful
-     * if you, for example, have a label that displays the width and
-     * height of the pixbuf in the view. 
+     * image view shows is changed and when its image data is changed.
+     * Listening to this signal is useful if you, for example, have a
+     * label that displays the width and height of the pixbuf in the
+     * view.
      *
      * <informalexample>
      *   <programlisting>
@@ -1249,11 +1254,11 @@ gtk_image_view_class_init (GtkImageViewClass *klass)
 
     /* Set zoom. */
     gtk_binding_entry_add_signal (binding_set, GDK_1, 0,
-                                  "set_zoom", 1, GTK_TYPE_DOUBLE, 1.0);
+                                  "set_zoom", 1, G_TYPE_DOUBLE, 1.0);
     gtk_binding_entry_add_signal (binding_set, GDK_2, 0,
-                                  "set_zoom", 1, GTK_TYPE_DOUBLE, 2.0);
+                                  "set_zoom", 1, G_TYPE_DOUBLE, 2.0);
     gtk_binding_entry_add_signal (binding_set, GDK_3, 0,
-                                  "set_zoom", 1, GTK_TYPE_DOUBLE, 3.0);
+                                  "set_zoom", 1, G_TYPE_DOUBLE, 3.0);
 
     /* Zoom in */
     gtk_binding_entry_add_signal (binding_set, GDK_plus, 0, "zoom_in", 0);
@@ -1269,7 +1274,7 @@ gtk_image_view_class_init (GtkImageViewClass *klass)
     /* Set fitting */
     gtk_binding_entry_add_signal (binding_set, GDK_x, 0,
                                   "set_fitting", 1,
-                                  GTK_TYPE_BOOL, TRUE);
+                                  G_TYPE_BOOLEAN, TRUE);
 
     /* Unmodified scrolling */
     gtk_binding_entry_add_signal (binding_set, GDK_Right, 0,
@@ -1436,6 +1441,47 @@ gtk_image_view_get_draw_rect (GtkImageView *view,
     return TRUE;
 }
 
+/**
+ * gtk_image_view_image_to_widget_rect:
+ * @view: a #GtkImageView
+ * @rect_in: a #GdkRectangle in image space coordinates to convert
+ * @rect_out: a #GdkRectangle to fill in with the widget space
+ *   coordinates
+ * @returns: %TRUE if the conversion was successful, %FALSE otherwise
+ *
+ * Convert a rectangle in image space coordinates to widget space
+ * coordinates. If the view is not realized, or if it contains no
+ * pixbuf, then the conversion was unsuccessful, %FALSE is returned
+ * and @rect_out is left unmodified.
+ *
+ * Note that this function may return a rectangle that is not visible
+ * on the widget.
+ **/
+gboolean
+gtk_image_view_image_to_widget_rect (GtkImageView *view,
+                                     GdkRectangle *rect_in,
+                                     GdkRectangle *rect_out)
+{
+    gdouble zoom = gtk_image_view_get_zoom (view);
+    GdkRectangle zoom_rect = {
+        rect_in->x * zoom,
+        rect_in->y * zoom,
+        rect_in->width * zoom,
+        rect_in->height * zoom
+    };
+
+    GdkRectangle image_rect, viewport;
+    if (!gtk_image_view_get_draw_rect (view, &image_rect))
+        return FALSE;
+    
+    gtk_image_view_get_viewport (view, &viewport);
+    rect_out->x = image_rect.x + zoom_rect.x - viewport.x;
+    rect_out->y = image_rect.y + zoom_rect.y - viewport.y;
+    rect_out->width = zoom_rect.width;
+    rect_out->height = zoom_rect.height;
+    return TRUE;
+}
+
 /**
  * gtk_image_view_get_check_colors:
  * @view: A #GtkImageView.
@@ -1476,7 +1522,7 @@ gtk_image_view_get_check_colors (GtkImageView *view,
  * the widget to repaint itself if it is realized.
  *
  * If @invalidate is %TRUE, the views entire area will be invalidated
- * instead of redrawn immidiately. The view is then queued for redraw,
+ * instead of redrawn immediately. The view is then queued for redraw,
  * which means that additional operations can be performed on it
  * before it is redrawn.
  *
@@ -1487,7 +1533,7 @@ gtk_image_view_get_check_colors (GtkImageView *view,
  * example.
  *
  * Normally, @invalidate should always be %FALSE because it is much
- * faster to repaint immidately than invalidating.
+ * faster to repaint immedately than invalidating.
  **/
 void
 gtk_image_view_set_offset (GtkImageView *view,
@@ -1513,7 +1559,7 @@ gtk_image_view_set_offset (GtkImageView *view,
  * then a grid with light and dark gray boxes will be drawn on the
  * transparent parts.
  *
- * Calling this method causes the widget to immidiately repaint. It
+ * Calling this method causes the widget to immediately repaint. It
  * also causes the ::pixbuf-changed signal to be emitted. This is done
  * so that other widgets (such as #GtkImageNav) will have a chance to
  * render a view of the pixbuf with the new transparency settings.
@@ -1573,7 +1619,7 @@ gtk_image_view_get_fitting (GtkImageView *view)
  * Sets whether to fit or not. If %TRUE, then the view will adapt the
  * zoom so that the whole pixbuf is visible.
  *
- * Setting the fitting causes the widget to immidiately repaint
+ * Setting the fitting causes the widget to immediately repaint
  * itself.
  *
  * Fitting is by default %TRUE.
@@ -1608,26 +1654,16 @@ gtk_image_view_get_pixbuf (GtkImageView *view)
  * @reset_fit: Whether to reset fitting or not.
  *
  * Sets the @pixbuf to display, or %NULL to not display any pixbuf.
- *
- * Normally, @reset_fit should be %TRUE which causes the fit mode to
- * be set to %GTK_FIT_SIZE_IF_LARGER. Which means that, initially, the
- * whole pixbuf will be shown.
+ * Normally, @reset_fit should be %TRUE which enables fitting. Which
+ * means that, initially, the whole pixbuf will be shown.
  *
  * Sometimes, the fit mode should not be reset. For example, if
  * GtkImageView is showing an animation, it would be bad to reset the
  * fit mode for each new frame. The parameter should then be %FALSE
  * which leaves the fit mode of the view untouched.
  *
- * This method must also be used to signal the view that the contents
- * of the pixbuf it display has been changed. This is the right way to
- * force the redraw:
- *
- * <informalexample>
- *   <programlisting>
- *     GdkPixbuf *pixbuf = gtk_image_view_get_pixbuf (GTK_IMAGE_VIEW (view));
- *     gtk_image_view_set_pixbuf (GTK_IMAGE_VIEW (view), pixbuf, FALSE);
- *   </programlisting>
- * </informalexample>
+ * This method should not be used if merely the contents of the pixbuf
+ * has changed. See gtk_image_view_damage_pixels() for that.
  *
  * If @reset_fit is %TRUE, the ::zoom-changed signal is emitted,
  * otherwise not. The ::pixbuf-changed signal is also emitted.
@@ -1666,7 +1702,7 @@ gtk_image_view_set_pixbuf (GtkImageView *view,
     
     g_signal_emit (G_OBJECT (view),
                    gtk_image_view_signals[PIXBUF_CHANGED], 0);
-    gtk_iimage_tool_pixbuf_changed (view->tool, reset_fit);
+    gtk_iimage_tool_pixbuf_changed (view->tool, reset_fit, NULL);
 }
 
 /**
@@ -1743,7 +1779,7 @@ gtk_image_view_get_black_bg (GtkImageView *view)
  *
  * Sets whether to draw a frame around the image or not. When %TRUE, a
  * one pixel wide frame is shown around the image. Setting this
- * attribute causes the widget to immidiately repaint itself.
+ * attribute causes the widget to immediately repaint itself.
  *
  * The default value is %TRUE.
  **/
@@ -1819,7 +1855,7 @@ gtk_image_view_get_show_cursor (GtkImageView *view)
  * the fastest, but provides bad rendering
  * quality. %GDK_INTERP_BILINEAR is a good compromise.
  *
- * Setting the interpolation mode causes the widget to immidiately
+ * Setting the interpolation mode causes the widget to immediately
  * repaint itself.
  *
  * The default interpolation mode is %GDK_INTERP_BILINEAR.
@@ -1857,7 +1893,7 @@ gtk_image_view_get_interpolation (GtkImageView *view)
  * gtk_iimage_tool_pixbuf_changed() is called so that the tool has a
  * chance to generate initial data for the pixbuf.
  *
- * Setting the tool causes the widget to immidiately repaint itself.
+ * Setting the tool causes the widget to immediately repaint itself.
  *
  * The default image tool is a #GtkImageToolDragger instance. See also
  * #GtkIImageTool.
@@ -1875,7 +1911,7 @@ gtk_image_view_set_tool (GtkImageView  *view,
     g_object_ref (view->tool);
 
     GtkWidget *widget = GTK_WIDGET (view);
-    gtk_iimage_tool_pixbuf_changed (tool, TRUE);
+    gtk_iimage_tool_pixbuf_changed (tool, TRUE, NULL);
     
     if (!GTK_WIDGET_REALIZED (widget))
         return;
@@ -1899,10 +1935,10 @@ gtk_image_view_get_tool (GtkImageView *view)
 /*************************************************************/
 /**
  * gtk_image_view_zoom_in:
- * @view: A #GtkImageView
+ * @view: a #GtkImageView
  *
  * Zoom in the view one step. Calling this method causes the widget to
- * immidiately repaint itself.
+ * immediately repaint itself.
  **/
 void
 gtk_image_view_zoom_in (GtkImageView *view)
@@ -1913,10 +1949,10 @@ gtk_image_view_zoom_in (GtkImageView *view)
 
 /**
  * gtk_image_view_zoom_out:
- * @view: A #GtkImageView
+ * @view: a #GtkImageView
  *
  * Zoom out the view one step. Calling this method causes the widget to
- * immidiately repaint itself.
+ * immediately repaint itself.
  **/
 void
 gtk_image_view_zoom_out (GtkImageView *view)
@@ -1924,3 +1960,74 @@ gtk_image_view_zoom_out (GtkImageView *view)
     gdouble zoom = gtk_zooms_get_zoom_out (view->zoom);
     gtk_image_view_set_zoom (view, zoom);
 }
+
+/**
+ * gtk_image_view_damage_pixels:
+ * @view: a #GtkImageView
+ * @rect: rectangle in image space coordinates to mark as damaged or
+ *   %NULL, to mark the whole pixbuf as damaged.
+ *
+ * Mark the pixels in the rectangle as damaged. That the pixels are
+ * damaged, means that they have been modified and that the view must
+ * redraw them to ensure that the visible part of the image
+ * corresponds to the pixels in that image. Calling this method emits
+ * the ::pixbuf-changed signal.
+ *
+ * This method must be used when <emphasis>modifying</emphasis> the
+ * image data:
+ *
+ * <informalexample>
+ *   <programlisting>
+ *     // Drawing something cool in the area 20,20 - 60,60 here...
+ *     ...
+ *     // And force an update
+ *     gtk_image_view_damage_pixels (view, &(GdkRectangle){20, 20, 60, 60});
+ *   </programlisting>
+ * </informalexample>
+ *
+ * If the whole pixbuf has been modified then @rect should be %NULL to
+ * indicate that a total update is needed.
+ *
+ * See also gtk_image_view_set_pixbuf().
+ **/
+void
+gtk_image_view_damage_pixels(GtkImageView *view,
+                             GdkRectangle *rect)
+{
+    g_signal_emit (G_OBJECT (view),
+                   gtk_image_view_signals[PIXBUF_CHANGED], 0);
+    gtk_iimage_tool_pixbuf_changed (view->tool, FALSE, rect);
+
+    if (rect)
+    {
+        // Convert rect to widget coordinates.
+        GdkRectangle wid_rect = {0};
+        gtk_image_view_image_to_widget_rect (view, rect, &wid_rect);
+
+        // The conversion from image to widget space coordinates is
+        // very inaccurate which is why a much larger area than the
+        // damaged has to be drawn. See the "Zoom bug" chapter in
+        // GtkImageToolSelector for more information.
+        gtk_widget_queue_draw_area (GTK_WIDGET (view),
+                                    wid_rect.x - wid_rect.width * 2,
+                                    wid_rect.y - wid_rect.height * 2,
+                                    wid_rect.width * 4, wid_rect.height * 4);
+    }
+    else
+        gtk_widget_queue_draw (GTK_WIDGET (view));
+}
+
+/**
+ * gtk_image_view_library_version:
+ * @returns: a string describing the version of GtkImageView.The
+ *   returned string is owned by GtkImageView and should not be
+ *   modified or freed.
+ *
+ * Returns a string with the format "major.minor.micro" which denotes
+ * the runtime version of GtkImageView being used.
+ **/
+const char *
+gtk_image_view_library_version (void)
+{
+    return PACKAGE_VERSION;
+}
